
"use strict";

var assert = require('assert');
var ast_visitor = require('./ast_visitor');

const Hole = ['HOLE'];
function isHole(val) {
    return val === Hole;
}

function print_depth(depth, s) {
    let spaces = "";
    for (var i = 0; i < depth; i++) {
        spaces += "  ";
    }
    console.log(spaces + s);
}

// A template object describes a partially specified tree (where the
// partial element may be empty, and the entire tree is fully specified).
//
// A AST subtree implicitly maps to a template object that matches
// that specific subtree.
//
// The set of template objects can be defined inductively:
//  An AST tree is a template tree if:
//      All of its fields have regular values, or a Hole.
//      All of its child nodes refer to template trees, or are Holes.
class Template {
    constructor(type, fields, children) {
        this.type = type;
        this.fields = fields;
        this.children = children;
    }

    equals(t) {
        if (t === Hole) {
            return false;
        }
        for (let f in this.fields) {
            if (!(f in t.fields))
                return false;
            if (this.fields[f] !== t.fields[f]) {
                return false;
            }
        }
        for (let f in t.fields) {
            if (!(f in this.fields))
                return false;
        }
        for (let c in this.children) {
            if (!(c in t.children))
                return false;
            if ((this.children[c] === Hole) && (t.children[c] === Hole))
                continue;
            if ((this.children[c] === Hole) && (t.children[c] !== Hole))
                return false;
            if ((this.children[c] !== Hole) && (t.children[c] === Hole))
                return false;

            if (!this.children[c].equals(t.children[c]))
                return false;
        }
        for (let c in t.children) {
            if (!(c in this.children))
                return false;
        }
        return true;
    }

    matches(t) {
        if (t === Hole) {
            return false;
        }
        for (let f in this.fields) {
            if (!(f in t.fields))
                return false;
            if (isHole(this.fields[f]))
                continue;
            if (this.fields[f] !== t.fields[f])
                return false;
        }
        for (let f in t.fields) {
            if (!(f in this.fields))
                return false;
        }
        for (let c in this.children) {
            if (!(c in t.children))
                return false;
            if (isHole(this.children[c]))
                continue;
            assert(!isHole(t.children[c]));
            if (!this.children[c].matches(t.children[c]))
                return false;
        }
        for (let c in t.children) {
            if (!(c in this.children))
                return false;
        }
        return true;
    }

    shallow_str() {
        let o = {};
        for (let n in this.fields) {
            o[n] = this.fields[n];
            if (o[n] === Hole) {
                o[n] = "<HOLE>";
            }
        }
        return this.type + " " + JSON.stringify(o);
    }

    print(depth) {
        if (!depth) { depth = 0; }
        print_depth(depth, this.shallow_str());
        for (let n in this.children) {
            if (this.children[n] === Hole) {
                print_depth(depth, "." + n + " => HOLE");
            } else {
                print_depth(depth, "." + n + " =>");
                this.children[n].print(depth + 1);
            }
        }
    }

    // Do a post-order iteration on the template tree.
    iter_post_order(f, counter) {
        if (!counter) {
            counter = {count: 0};
        }
        for (let c in this.children) {
            if (this.children[c] !== Hole) {
                this.children[c].iter_post_order(f, counter);
            }
        }
        f(this, counter.count);
        counter.count += 1;
    }

    weight() {
        return Template.calculate_weight(this);
    }

    static calculate_weight(template) {
        if (template === Hole)
            return 0;

        let w = 1;
        for (let f in template.fields) {
            if (template.fields[f] !== Hole) {
                w += 1;
            }
        }
        for (let c in template.children) {
            if (template.children[c] !== Hole) {
                w += Template.calculate_weight(template.children[c]);
            }
        }
        return w;
    }
}


// Given two templates, construct a new template object that matches both.
function generate_template(tree1, tree2) {
    // If the types don't match, the only template is a hole
    if (tree1.type != tree2.type) {
        return Hole;
    }

    // Otherwise, compare the fields on each.
    let seen_fields = {};
    let result = new Template(tree1.type, {}, {});

    function add_template_field(name, v1, v2) {
        if (v1 === undefined && v2 === undefined) {
            // Both are not given, skip property.
            return;
        }
        result.fields[name] = (v1 === v2) ? v1 : Hole;
    }

    for (let f1 in tree1.fields) {
        seen_fields[f1] = true;
        add_template_field(f1, tree1.fields[f1], tree2.fields[f1]);
    }
    for (let f2 in tree2.fields) {
        if (seen_fields[f2]) { continue; }
        add_template_field(f2, tree1.fields[f2], tree2.fields[f2]);
    }

    // Generate a template object for each child.
    let seen_children = {};

    function add_template_child(name, c1, c2) {
        if (c1 === undefined && c2 === undefined) {
            // Both are not given, skip property.
            return;
        }
        if (c1 === undefined || c2 === undefined) {
            // One not given, replace with Hole
            result[name] = Hole;
            return;
        }
        result.children[name] = generate_template(c1, c2);
    }
    for (let n1 in tree1.children) {
        seen_children[n1] = true;
        add_template_child(n1, tree1.children[n1], tree2.children[n1]);
    }
    for (let n2 in tree2.children) {
        if (seen_children[n2]) { continue; }
        add_template_child(n2, tree1.children[n2], tree2.children[n2]);
    }

    return result;
}

function ast_to_template(ast) {
    let v = new ast_visitor.Visitor(handler);
    v.visit(ast);
    return ast.template;

    function handler(handlerName, {ast, parent, name}, fields, children) {
        if (handlerName.endsWith("End")) { return; }

        // Make a new template object with empty children.
        let t = new Template(ast.type, fields, {});

        // Add it as a child to any parent node.
        if (parent) {
            parent.template.children[name] = t;
        }

        // Attach it to this ast node.
        ast.template = t;
    }
}

module.exports = {
    Template, ast_to_template, generate_template, isHole
};
