
"use strict";

var assert = require('assert');
var fs = require('fs');
var crypto = require('crypto');

function readFile(filename) {
    return fs.readFileSync(filename, 'utf8');
}

function DEBUG(str) {
    console.log(str);
}

function makeShallow(obj) {
    if (typeof(obj) != 'object') {
        return obj;
    }
    var names = Object.getOwnPropertyNames(obj);
    var result = {};
    for (let name of names) {
        let val = obj[name];
        let tv = typeof(val);
        if (tv === 'object') {
            if (val instanceof Array) {
                val = "[Array]"
            } else {
                val = "[Object]"
            }
        } else if(tv === 'function') {
            val = "[Function]"
        }
        result[name] = val;
    }
    return result;
}
function shallowStringify(obj) {
    return JSON.stringify(makeShallow(obj));
}

function hashObject(obj, h) {
    if (typeof(h) == 'string') {
        h = new crypto.Hash(h);
    } else if (typeof('h') != 'object') {
        h = new crypto.Hash('SHA256');
    }

    if (typeof(obj) != 'object') {
        h.write(JSON.stringify(obj));
        return h;
    }

    let keys = Object.getOwnPropertyNames(obj);
    keys.sort();
    for (let k of key) {
        let ks = k.toString();
        let vs = hashObject(obj, h).digest('UTF8');
        h.write(ks.length + "|" + ks + vs.length + "|" + vs);
    }
    return h;
}

class VisitorCase {
    constructor({visitor, type, fields, child_fields}) {
        this.visitor = visitor;
        this.type = type;
        this.fields = fields;
        this.child_fields = child_fields;
        this.known_fields = new Set();

        assert(this.fields.indexOf('type') < 0, "Fields descriptor should not contain 'type'");

        for (let f of this.fields) {
            this._normalizeName(f, (kind, nm) => this.known_fields.add(nm));
        }
        for (let f of this.child_fields) {
            this._normalizeName(f, (kind, nm) => this.known_fields.add(nm));
        }
    }

    toString() {
        return "[VisitorCase " + this.type +
               " fields=" + this.fields.join(',') +
               " child_fields=" + this.child_fields.join(',') + "]"
    }

    edgeString(edge) {
        return this.type + ":" + edge;
    }

    visitCase(ast, parent, name, f, fend) {
        if (false && KVKV && this.fields.length == 0) {
            DEBUG("Case: " + this.toString());
            DEBUG("VISITING: " + shallowStringify(ast));
        }
        assert.equal(ast.type, this.type, "ast.type == this.type");

        // Confirm that the ast node conforms to our expectations.
        for (let astField of Object.getOwnPropertyNames(ast)) {
            assert((astField == 'type') || this.known_fields.has(astField),
                   "Unknown field " + this.edgeString(astField) + " found.");
        }

        let params = {ast, parent, name};
        let data = {};
        for (let field of this.fields) {
            this._collectDataField(ast, field, data);
        }
        let children = {};
        for (let field of this.child_fields) {
            this._collectChildField(ast, field, children);
        }

        if (f(params, data, children) === false) {
            return false;
        }
        for (let field of this.child_fields) {
            this._dispatchChildField(ast, field);
        }
        fend(params, data, children);
    }

    _normalizeName(name, callback) {
        if (name.startsWith("*")) {
            return callback('array', name.substr(1));
        } else if (name.startsWith("?")) {
            return callback('optional', name.substr(1));
        } else if (name.startsWith("!")) {
            return callback('ignore', name.substr(1));
        } else {
            return callback('plain', name);
        }
    }

    _collectDataField(ast, edge, data) {
        this._normalizeName(edge, (kind, name) => {
            if (kind == 'optional') {
                if (ast[edge]) {
                    data[edge] = ast[edge];
                }
            } else if (kind == 'plain') {
                assert(ast[edge] !== undefined, "Required parameter " + this.type + "." + edge + " not given.");
                data[edge] = ast[edge];
            }
        });
    }

    _collectChildField(ast, edge, children) {
        this._normalizeName(edge, (kind, name) => {
            if (kind == 'optional') {
                if (ast[edge]) {
                    children[edge] = ast[edge];
                }
            } else if (kind == 'plain') {
                assert(ast[edge] !== undefined, "Required child " + this.type + "::" + edge + " not given.");
                children[edge] = ast[edge];
            }
        });
    }

    _dispatchChildField(ast, edge) {
        this._normalizeName(edge, (kind, name) => {
            if (kind == 'array') {
                this._dispatchArrayChildren(ast, name);
            } else if (kind == 'optional' || kind == 'plain') {
                this._dispatchSingleChild(ast, name, kind == 'optional');
            } else {
                assert(false, "Invalid child edge kind: " + kind);
            }
        });
    }
    _dispatchArrayChildren(ast, name) {
        // Array of children.
        const arr = ast[name];
        for (var i = 0; i < arr.length; i++) {
            if (this.visitor.visitNode(arr[i], ast, name+"."+i) === false) {
                return false;
            }
        }
    }
    _dispatchSingleChild(ast, name, optional) {
        if (!ast[name]) {
            assert(optional, "Non-optional child name " + this.edgeString(name) + " was not found.");
            return;
        }
        if (this.visitor.visitNode(ast[name], ast, name) === false) {
            return false;
        }
    }

    static splitFields(fields) {
        return fields == "null" ? [] : fields.split(',');
    }

    static fromString(visitor, str) {
        // split on space.
        let [type, fields, child_fields] = str.split(/ +/);
        fields = VisitorCase.splitFields(fields)
        child_fields = VisitorCase.splitFields(child_fields)
        return new VisitorCase({visitor, type, fields, child_fields});
    }

    static fromStringArray(visitor, strarr) {
        let result = {};
        for (let str of strarr) {
            let vc = VisitorCase.fromString(visitor, str);
            result[vc.type] = vc;
        }
        return result;
    }
}

const AST_NODES = [
    "Program                        sourceType                          *body",
    "ClassDeclaration               null                                id,?superClass,body",
    "ClassBody                      null                                *body",
    "MethodDefinition               computed,kind,static                key,value",
    "ObjectPattern                  null                                *properties",
    "Property                       computed,kind,method,shorthand      key,value",
    "FunctionDeclaration            generator,expression,async          id,*params,body",
    "VariableDeclaration            kind                                *declarations",
    "VariableDeclarator             null                                id,?init",
    "EmptyStatement                 null                                null",
    "BlockStatement                 null                                *body",
    "ThrowStatement                 null                                argument",
    "ReturnStatement                null                                ?argument",
    "IfStatement                    null                                test,consequent,?alternate",
    "ForStatement                   null                                ?init,?test,?update,body",
    "DoWhileStatement               null                                body,test", 
    "WhileStatement                 null                                test,body",
    "ForOfStatement                 null                                left,right,body",
    "ForInStatement                 each                                left,right,body",
    "ContinueStatement              null                                ?label",
    "BreakStatement                 null                                ?label",
    "TryStatement                   null                                block,handler,?finalizer",
    "CatchClause                    null                                param,body",
    "ExpressionStatement            ?directive                          expression",
    "AssignmentExpression           operator                            left,right",
    "ConditionalExpression          null                                test,consequent,alternate",
    "SequenceExpression             null                                *expressions",
    "LogicalExpression              operator                            left,right",
    "BinaryExpression               operator                            left,right",
    "UnaryExpression                operator,prefix                     argument",
    "UpdateExpression               operator,prefix                     argument",
    "CallExpression                 null                                callee,*arguments",
    "NewExpression                  null                                callee,*arguments",
    "MemberExpression               computed                            object,property",
    "FunctionExpression             generator,expression,async          ?id,*params,body",
    "ArrowFunctionExpression        generator,expression,async          ?id,*params,body",
    "ObjectExpression               null                                *properties",
    "ArrayExpression                null                                *elements",
    "ArrayPattern                   null                                *elements",
    "Literal                        value,!raw,?regex                   null",
    "Identifier                     name                                null",
    "ThisExpression                 null                                null",
    "Super                          null                                null",
];

class Visitor {
    constructor(target) {
        this.cases_ = VisitorCase.fromStringArray(this, AST_NODES);
        this.target_ = target || this;
        if (typeof(this.target_) == "function") {
            this.dispatcher_ = (methodName, params, data, children) => {
                this.target_(methodName, params, data, children);
            };
        } else {
            this.dispatcher_ = (methodName, params, data, children) => {
                if (this.target_[methodName]) {
                    return this.target_[methodName](params, data, children);
                } else {
                    if (!methodName.endsWith("End")) {
                        DEBUG("!!! WARNING: FAILED TO DISPATCH visitor method `" + methodName + "` !!!");
                    }
                }
            };
        }
    }

    visit(ast) {
        return this.visitNode(ast);
    }

    visitNode(ast, parent, name) {
        let c = this.cases_[ast.type];
        if (!c) {
            throw new Error("Unhandled node: " + shallowStringify(ast));
        }
        return c.visitCase(ast, parent, name,
                           (params,data,children) => this.dispatcher_(params.ast.type, params, data, children),
                           (params,data,children) => this.dispatcher_(params.ast.type + "End", params, data, children));
    }
}

class PrintVisitor extends Visitor {
    constructor() {
        super();
        this.indentLevel_ = 0;
        this.indentPrefix_ = "";
    }

    indent() {
        this.indentLevel_ += 1;
        this.indentPrefix_ += "  ";
    }
    dedent() {
        this.indentLevel_ -= 1;
        assert(this.indentLevel_ >= 0, "this.indentLevel_ >= 0");
        this.indentPrefix_ = this.indentPrefix_.substr(2);
    }
    indentString(str) {
        //return this.indentPrefix_ + "[" + this.indentLevel_ + "] " + str;
        return this.indentPrefix_ + str;
    }


    log_raw(str) {
        console.log(str);
    }
    log(str) {
        this.log_raw(this.indentString(str));
    }

    log_begin(name, ast, data) {
        var s = ast.type + "(" + shallowStringify(data) + ") {";
        if (name !== undefined) {
            s = name + " => " + s;
        }
        this.log(s);
    }
    log_end(ast) {
        this.log("} //" + ast.type);
    }

    Program({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    ProgramEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    ClassDeclaration({ast, name}, data) {
        assert.equal(ast.id.type, "Identifier");
        this.log_begin(name, ast, data);
        this.indent();
    }
    ClassDeclarationEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    ClassBody({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    ClassBodyEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    MethodDefinition({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    MethodDefinitionEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    ObjectPattern({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    ObjectPatternEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    Property({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    PropertyEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    FunctionDeclaration({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    FunctionDeclarationEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    VariableDeclaration({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    VariableDeclarationEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    VariableDeclarator({ast, name}, data) {
        assert((ast.id.type == "Identifier") || (ast.id.type == "ArrayPattern"));
        this.log_begin(name, ast, data);
        this.indent();
    }
    VariableDeclaratorEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    BlockStatement({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    BlockStatementEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    ForStatement({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    ForStatementEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    ForOfStatement({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    ForOfStatementEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    IfStatement({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    IfStatementEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    ReturnStatement({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    ReturnStatementEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    ExpressionStatement({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    ExpressionStatementEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    AssignmentExpression({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    AssignmentExpressionEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    ConditionalExpression({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    ConditionalExpressionEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    LogicalExpression({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    LogicalExpressionEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    BinaryExpression({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    BinaryExpressionEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    UnaryExpression({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    UnaryExpressionEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    UpdateExpression({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    UpdateExpressionEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    CallExpression({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    CallExpressionEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    NewExpression({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    NewExpressionEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    MemberExpression({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    MemberExpressionEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    FunctionExpression({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    FunctionExpressionEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    ArrowFunctionExpression({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    ArrowFunctionExpressionEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    ObjectExpression({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    ObjectExpressionEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    ArrayExpression({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    ArrayExpressionEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    ArrayPattern({ast, name}, data) {
        this.log_begin(name, ast, data);
        this.indent();
    }
    ArrayPatternEnd({ast}) {
        this.dedent();
        this.log_end(ast);
    }

    Literal({ast, name}, data) {
        this.log(name + " => Literal(" + typeof(data.value) + ":" + JSON.stringify(data.value) + ")");
    }

    Identifier({ast, name}, data) {
        this.log(name + " => Ident(" + data.name + ")");
    }

    ThisExpression({ast, name}, data) {
        this.log(name + " => This");
    }
}

module.exports = {
    readFile, Visitor, PrintVisitor
};
